"""AvroToSQL manager."""

from popsink_transformer.errors import PopsinkTransformerError
from popsink_transformer.avro_to_sql.utils.type_converter import (
    SqlEngine,
    get_avro_to_sql_mapping,
)


class AvroToSQL:
    """AvroToSQL class.

    Note: This class is designed to convert Avro fields into SQL-readable types.
    It does not perform the conversion of entire Avro schemas to SQL queries.
    """

    def __init__(self, from_source: dict, sql_engine: SqlEngine) -> None:
        self.from_source = from_source
        self.sql_engine = sql_engine

    def convert(self) -> str:
        """Convert an avro schema to an SQL schema."""
        self._handle_base_schema_errors()

        sql_fields = []
        for avro_field in self.from_source["fields"]:
            formatted_field = self._format_field(avro_field)
            sql_fields.append(formatted_field)

        sql_schema = ",".join(sql_fields)
        return sql_schema

    def _format_field(self, avro_field: dict) -> dict:
        """Call all the specific methods for field formatting."""
        sql_field = ""
        sql_field = self._unwrap_field(avro_field, sql_field)
        sql_field = self._write_nullable(avro_field, sql_field)
        sql_field = self._format_array(avro_field, sql_field)
        sql_field = self._convert_avro_type_to_sql_type(sql_field)
        sql_field = self._format_record(avro_field, sql_field)
        sql_field = self._format_map(avro_field, sql_field)
        return sql_field

    def _unwrap_field(self, avro_field: dict, sql_field: str):
        """Unwrap avro field."""
        avro_field_type = avro_field["type"]

        if avro_field.get("logicalType"):
            avro_field_type = avro_field["logicalType"]

        if isinstance(avro_field_type, list):
            avro_field_type = [
                value for value in avro_field_type if value != "null"
            ].pop()

        if isinstance(avro_field_type, dict):
            avro_field_type = avro_field_type["type"]

        sql_field = f'`{avro_field["name"]}` {avro_field_type}'
        return sql_field

    def _write_nullable(self, avro_field: dict, sql_field: str):
        """Write the nullable state at the end of the field."""
        if "type" in avro_field:
            avro_field_type = avro_field["type"]

            if isinstance(avro_field_type, dict):
                avro_field_type = avro_field_type.get("type")

            if avro_field_type != "record":
                if (
                    not isinstance(avro_field["type"], list)
                    and avro_field["type"] != "record"
                ):
                    sql_field += " NOT NULL"
        return sql_field

    def _format_array(self, avro_field: dict, sql_field: str):
        """Format array fields."""
        if "array" in sql_field:
            if self.sql_engine is SqlEngine.POSTGRES:
                sql_field = sql_field.replace(
                    "array", f"{avro_field['type']['items']} ARRAY"
                )
            if self.sql_engine is SqlEngine.FLINK:
                sql_field = sql_field.replace(
                    "array", f"ARRAY<{avro_field['type']['items']}>"
                )
        return sql_field

    def _format_map(self, avro_field: dict, sql_field: str):
        """Format map fields."""
        if "map" in sql_field:
            if self.sql_engine is SqlEngine.POSTGRES:
                sql_field = sql_field.replace("map", "JSON")

            if self.sql_engine is SqlEngine.FLINK:
                map_type = avro_field["type"]["values"].upper()
                sql_field = sql_field.replace(
                    "map", f"MAP<{map_type}, {map_type}>"
                )
        return sql_field

    def _format_record(self, avro_field: dict, sql_field: str):
        """Format record fields."""
        if "record" in sql_field:
            if self.sql_engine is SqlEngine.POSTGRES:
                sql_field = sql_field.replace("record", "JSON")

            if self.sql_engine is SqlEngine.FLINK:
                nested_fields = avro_field.get(
                    "fields", avro_field["type"]["fields"]
                )

                field_header = []
                for field in nested_fields:
                    formatted_field = self._format_field(field)
                    field_header.append(formatted_field)

                header = ", ".join(field_header)
                sql_field = sql_field.replace("record", f"ROW<{header}>")

                field_body = []
                record_name = avro_field["name"]
                for field in nested_fields:
                    field_name = field["name"]
                    line = f"`{field_name}` AS `{record_name}`.`{field_name}`"
                    field_body.append(line)

                body = ",".join(field_body)
                sql_field += f",{body}"
        return sql_field

    def _convert_avro_type_to_sql_type(self, sql_field: str):
        """Convert avro type to SQL type.

        We replace the avro type with one space before to not mistake it with
        the name of the field.

        Example: '`date` date'
        """

        for avro_type, sql_type in get_avro_to_sql_mapping(
            self.sql_engine
        ).items():
            if avro_type in sql_field:
                sql_field = sql_field.replace(f" {avro_type}", f" {sql_type}")
                if self.sql_engine is SqlEngine.FLINK:
                    sql_field = sql_field.replace(
                        f"<{avro_type}>", f"<{sql_type}>"
                    )
        return sql_field

    def _handle_base_schema_errors(self) -> None:
        """Handle base schema errors."""
        self._raise_error_if_schema_is_empty()
        self._raise_error_if_no_record()
        self._raise_error_if_no_fields_key()
        self._raise_error_if_fields_key_is_not_a_list()

    def _raise_error_if_no_record(self) -> None:
        """Raise error if 'record' type is not found."""
        if self.from_source.get("type") != "record":
            raise PopsinkTransformerError("The base type should be a 'record'")

    def _raise_error_if_no_fields_key(self) -> None:
        """Raise error if 'fields' key is missing."""
        if "fields" not in self.from_source:
            raise PopsinkTransformerError(
                "Missing 'fields' key on base schema"
            )

    def _raise_error_if_schema_is_empty(self) -> None:
        """Raise error if the avro schema is empty."""
        if not self.from_source:
            raise PopsinkTransformerError(
                "Expecting value: line 1 column 1 (char 0)"
            )

    def _raise_error_if_fields_key_is_not_a_list(self) -> None:
        """Raise error if the 'fields' key is not a list."""
        if not isinstance(self.from_source.get("fields"), list):
            raise PopsinkTransformerError("The 'fields' key should be a list")



"""AvroToJson manager."""


class AvroToJson:
    """AvroToJson class.

    Fixed and bytes fields are not yet fully compatible and are converted to string type.
    """

    AVRO_TO_JSON = {
        "int": "integer",
        "long": "integer",
        "float": "number",
        "double": "number",
        "record": "struct",
        "enum": "string",
        "bytes": "string",
        "map": "object",
        "fixed": "string",
    }

    def __init__(self, from_source: dict) -> None:
        self.from_source = from_source

    def convert(self) -> dict:
        """Convert an avro schema to a json schema."""
        json_schema = self._get_base_structure()

        if not self.from_source:
            return json_schema

        json_fields = []
        for field in self.from_source["fields"]:
            formatted_field = self._format_field(field)
            json_fields.append(formatted_field)

        json_schema["schema"]["fields"] = json_fields
        return json_schema

    def _convert_avro_type_to_json_type(self, avro_type: str):
        """Convert avro type to json type."""
        return self.AVRO_TO_JSON.get(avro_type, avro_type)

    def _format_field(self, field: dict) -> dict:
        """Call all the specific methods for field formatting."""
        json_field = {}
        json_field = self._write_optional_state(field, json_field)
        field = self._unwrap_complex_type(field)
        json_field = self._add_type(field, json_field)
        json_field = self._format_array(field, json_field)
        json_field = self._format_record(field, json_field)
        json_field = self._format_enum(field, json_field)
        json_field = self._format_map(field, json_field)
        json_field = self._format_uuid(field, json_field)
        json_field = self._format_date(field, json_field)
        json_field = self._convert_name_to_field(field, json_field)
        json_field["type"] = self._convert_avro_type_to_json_type(
            json_field["type"]
        )
        return json_field

    def _add_type(self, field: dict, json_field: dict):
        """Add the type."""
        json_field["type"] = field["type"]
        return json_field

    def _unwrap_complex_type(self, field: dict) -> dict:
        """Unwrap the complexe type if it is found."""
        if isinstance(field["type"], dict):
            wrapped_field: dict = field["type"]
            if "name" not in wrapped_field:
                wrapped_field["name"] = field["name"]
            return wrapped_field
        return field

    def _get_base_structure(self) -> dict:
        """Return json schema base structure."""
        base_structure = {
            "schema": {
                "type": "struct",
                "optional": False,
                "fields": [],
            }
        }
        return base_structure

    def _convert_name_to_field(self, field: dict, json_field: dict) -> dict:
        """Convert key 'name' to 'field'."""
        json_field["field"] = field["name"]
        return json_field

    def _write_optional_state(self, field: dict, json_field: dict) -> dict:
        """Write the correct optional state - bool."""
        is_optional = False
        if isinstance(field["type"], list):
            real_type = [item for item in field["type"] if item != "null"][0]
            field["type"] = real_type
            is_optional = True
        json_field["optional"] = is_optional
        return json_field

    def _format_array(self, field: dict, json_field: dict) -> dict:
        """Format array fields."""
        if field["type"] == "array":
            json_field["items"] = {
                "type": self._convert_avro_type_to_json_type(field["items"])
            }
        return json_field

    def _format_record(self, field: dict, json_field: dict) -> dict:
        """Format record fields."""
        if field["type"] == "record":
            json_fields = []
            for inner_field in field["fields"]:
                formatted_field = self._format_field(inner_field)
                json_fields.append(formatted_field)
            json_field["fields"] = json_fields
        return json_field

    def _format_enum(self, field: dict, json_field: dict) -> dict:
        """Format enum fields."""
        if field["type"] == "enum":
            json_field["enum"] = field["symbols"]
        return json_field

    def _format_map(self, field: dict, json_field: dict) -> dict:
        """Format map fields."""
        if field["type"] == "map":
            json_field["patternProperties"] = {
                "*": {
                    "type": self._convert_avro_type_to_json_type(
                        field["values"]
                    )
                }
            }
        return json_field

    def _format_uuid(self, field: dict, json_field: dict) -> dict:
        """Format uuid fields."""
        if field.get("logicalType") == "uuid":
            json_field["format"] = field["logicalType"]
        return json_field

    def _format_date(self, field: dict, json_field: dict) -> dict:
        """Format date fields."""
        if field.get("logicalType") == "date":
            json_field["format"] = field["logicalType"]
            json_field["type"] = "string"
        return json_field



"""Transformer manager."""

from popsink_transformer.avro_to_json.main import AvroToJson
from popsink_transformer.avro_to_sql.main import AvroToSQL, SqlEngine


class Transformer:
    """Transformer class."""

    def __init__(
        self,
        from_source: dict = None,
        sql_engine: SqlEngine = None,
    ) -> None:
        self.avro_to_json = AvroToJson(from_source=from_source)
        self.avro_to_sql = AvroToSQL(
            from_source=from_source,
            sql_engine=sql_engine,
        )



"""Type converter."""

from enum import Enum
from popsink_transformer.errors import PopsinkTransformerError


class SqlEngine(Enum):
    """SQL types supported."""

    POSTGRES = "postgres"
    FLINK = "flink"


def get_avro_to_sql_mapping(sql_engine: SqlEngine):
    """Get the appropriate Avro to SQL mapping based on the SQL type."""
    if sql_engine is SqlEngine.POSTGRES:
        return {
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE PRECISION",
            "bytes": "BIT",
            "string": "STRING",
            "enum": "JSON",
            "fixed": "JSON",
            "uuid": "UUID",
            "date": "DATE",
        }

    if sql_engine is SqlEngine.FLINK:
        return {
            "boolean": "BOOLEAN",
            "int": "INTEGER",
            "long": "BIGINT",
            "float": "FLOAT",
            "double": "DOUBLE",
            "bytes": "BYTES",
            "string": "STRING",
            "enum": "STRING",
            "fixed": "DECIMAL",
            "date": "DATE",
            "uuid": "STRING",
        }
    raise PopsinkTransformerError("Incorrect SQL type has been specified.")


def convert_avro_type_to_sql_type(
    avro_field_type: str, sql_engine: SqlEngine
) -> str:
    """Convert an Avro type to SQL type."""
    if not avro_field_type:
        raise PopsinkTransformerError("Avro field type cannot be empty")

    if not sql_engine:
        raise PopsinkTransformerError("SQL type cannot be empty")

    avro_to_sql_mapping = get_avro_to_sql_mapping(sql_engine=sql_engine)

    if avro_field_type not in avro_to_sql_mapping:
        raise PopsinkTransformerError(
            f"No matching SQL type found for Avro type: {avro_field_type}"
        )

    return avro_to_sql_mapping[avro_field_type]
